#!/usr/bin/python
################################################################################
#
# SOAP.py 0.9 - Cayce Ullman    (cayce@actzero.com)
#               Brian Matthews  (blm@actzero.com)
#
# INCLUDED:
# - General SOAP Parser based on sax.xml (requires Python 2.0)
# - General SOAP Builder
# - SOAP Proxy for RPC client code
# - SOAP Server framework for RPC server code
#
# FEATURES:
# - Handles all of the types in the BDG
# - Handles faults
# - Allows namespace specification
# - Allows SOAPAction specification
# - Homogeneous typed arrays
# - Supports multiple schemas
# - Header support (mustUnderstand and actor)
# - XML attribute support
# - Multi-referencing support (Parser/Builder)
# - Good interop, passes all client tests for Frontier, SOAP::LITE, SOAPRMI
# - Encodings
#
# TODO:
# - More date types - BLM
# - Arrays (sparse, multidimensional and partial) - BLM
# - HTTPS - BLM
# - Type coercion system (Builder) - MCU
# - Early WSDL Support - MCU
# - Determine sequences on parsing? - MCU
# - Type by name - MCU
# - SOAP:ENC root - MCU
# - Improve attribute interface - MCU
# - setup.py - MCU
# - Documentation
# - Look at performance
#
################################################################################
#
# Copyright (c) 2001, actzero, inc.
# Copyright (c) 2001, Cayce Ullman.
# Copyright (c) 2001, Brian Matthews.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# Neither the name of actzero, inc. nor the names of its contributors may
# be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
################################################################################
import xml.sax
import xml.sax.saxutils as saxutils
import UserList
import base64
import exceptions
import re
import sys
import time
import SocketServer
from types import *

ident = '$Id: SOAP.py,v 1.193 2001/05/01 21:22:28 cayce Exp $'

DEBUG = 0
__version__ = "0.9"

# Platform hackery

# Check float support
try:
    float("NaN")
    float("INF")
    float("-INF")
    good_float = 1
except:
    good_float = 0

################################################################################
# Exceptions
################################################################################
class Error(exceptions.Exception):
    def __init__(self, msg):
        self.msg = msg
    def __str__(self):
        return "<Error : %s>" % self.msg
    def __repr__(self):
        return "<Error : %s>" % self.msg

class HTTPError(Error):
    # indicates an HTTP protocol error
    def __init__(self, code, msg):
        self.code = code
        self.msg  = msg
    def __repr__(self):
        return "<HTTPError %s %s>" % (self.code, self.msg)
    __str__=__repr__

##############################################################################
# Namespace Class
################################################################################
class NS:
    ENV  = "http://schemas.xmlsoap.org/soap/envelope/"
    ENC  = "http://schemas.xmlsoap.org/soap/encoding/"

    XSD  = "http://www.w3.org/1999/XMLSchema"

    XSD2 = "http://www.w3.org/2000/10/XMLSchema"
    XSD3 = "http://www.w3.org/2001/XMLSchema"

    XSD_L = [XSD,XSD2,XSD3]
    EXSD_L= [ENC,XSD,XSD2,XSD3]

    XSI   = "http://www.w3.org/1999/XMLSchema-instance"
    XSI2  = "http://www.w3.org/2000/10/XMLSchema-instance"
    XSI3  = "http://www.w3.org/2001/XMLSchema-instance"
    XSI_L = [XSI, XSI2, XSI3]

    #URN  = "http://www.xmethods.com/service"
    URN   = "http://soapinterop.org/xsd"

    # For generated messages
    ENV_T = "SOAP-ENV"
    ENC_T = "SOAP-ENC"
    XSD_T = "xsd"
    XSD2_T= "xsd2"
    XSD3_T= "xsd3"
    XSI_T = "xsi"
    URN_T = "urn"

    def __init__(self):
        raise Error, "Don't instantiate this"

################################################################################
# Types and Wrappers
################################################################################

# Anything that can have a name will end up being based on this.

class NamedInstance:
    def __init__(self, name = None):
        self._name = name

    def __repr__(self):
        if self._name:
            return "<%s %s at %d>" % (self.__class__,self._name, id(self))
        return "<%s at %d>" % (self.__class__, id(self))

class Void(NamedInstance):
    pass

class Compound(NamedInstance):
    def __init__(self, name = None):
        self._name     = name
        self._attr_map = {}   # For holding attributes
        self._attr_ord = {}
        self._refmap   = {}   # For holding references
        self._refmap_r = {}
        self._ns       = None
        self.__empty   = {}

    # Adds an XML attribute to a python object attribute
    def _setAttr(self, tag, attr, value):
        if type(tag) == IntType:
            try:
                self._attr_ord[tag][attr] = value
            except KeyError:
                self._attr_ord[tag] = {attr: value}
        else:
            try:
                self._attr_map[tag][attr] = value
            except KeyError:
                self._attr_map[tag] = {attr: value}

    def _getAttrs(self, attr):
        if type(attr) == IntType:
            try:
                return self._attr_ord[attr]
            except:
                return self.__empty

        if self._attr_map.has_key(attr):
            return self._attr_map[attr]
        else:
            return self.__empty

# This is a struct, it also is going to commonly be an actual method
class Struct(Compound):
    def __init__(self, name = None, data=None):
        Compound.__init__(self, name)
        self._aslist = []
        self._asdict = {}
        self._keyord = []

        if type(data)==DictType:
            self.__dict__.update(data)

    def __getitem__(self, item):
        if type(item) == IntType:
            return self._aslist[item]
        return getattr(self, item)

    def _keys(self):
        return filter(lambda x: x[0]!='_', self.__dict__.keys())

    def _addElem(self, name, value):
        self.__dict__[name] = value
        self._asdict[name]  = value  # for convenience
        self._keyord.append(name)
        self._aslist += (value,)

    def _getList(self, name):
        out = []
        count = 0
        for i in self._keyord:
            if i==name:
                out.append(self._aslist[count])
            count +=1
        return out

    # As in _getList return a list of the named elements found
    # in this Struct but also return a second list whose entries
    # correspond one to one with the element list and contain
    # the XML attribute dictionary for that element.
    def _getAttributedList(self, name):
        # Determine what the index is for that instance and
        # see if that instance has any attributes.

        data_list = []
        attr_list = []
        index = 0
        for index in range(len(self._keyord)):
            if self._keyord[index] == name:
                data_list.append( self._aslist[index] )
                try:
                    attr_list.append( self._attr_ord[index] )
                except:
                    attr_list.append( {} )
        return data_list, attr_list

class Header(Struct):
    def __init__(self, name = None, data=None):
         name = name or NS.ENV_T + ":Header"
         Struct.__init__(self, name, data)

    def _setMustUnderstand(self, tag, val):
        self._setAttr(tag, (NS.ENV, "mustUnderstand"), val)

    def _getMustUnderstand(self, tag):
        return self._getAttrs(tag)[(NS.ENV, "mustUnderstand")]

    def _setActor(self, tag, val):
        self._setAttr(tag, (NS.ENV, "actor"), val)

    def _getActor(self, tag):
        return self._getAttrs(tag)[(NS.ENV, "actor")]

    def __getitem__(self, item):
        return getattr (self, item)

class Body(Struct):
    def __init__(self, name = None, data = None):
        name = name or NS.ENV_T + ":Body"
        Struct.__init__(self, name, data)

class Array(UserList.UserList, Compound):
    def __init__(self, name = None, data = None):
        if data:
            if type(data) not in (ListType, TupleType):
                raise Error, "Data must be a sequence"

        UserList.UserList.__init__(self, data)
        Compound.__init__(self, name)

class TypedArray(Array):
    def __init__(self, name=None, array_type = None, data = []):
        self._type = array_type
        Array.__init__(self, name, data)

class Fault(Struct, Error):
    def __init__(self, faultcode="", faultstring="", detail=None):
        self.faultcode = faultcode
        self.faultstring = faultstring
        if detail:
            self.detail = detail
        Struct.__init__(self, "Fault")

    def __repr__(self):
        return "<Fault %s: %s>" % (self.faultcode, self.faultstring)

    __str__ = __repr__

class Boolean(NamedInstance):
    def __init__(self, data=0, name=None):
        self._data = not not int(data)
        self._name = name

    def __repr__(self):
        if self._name:
            return "<Boolean %s %d>" % (self._name, self._data)
        return "<Boolean %d>" % (self._data)

    def __int__(self):
        return self._data

    __nonzero__ = __int__

class DateTime(NamedInstance):
    def __init__(self, data = None, name = None):
        self._name = name

        frac = 0
        if data == None:
            data = time.localtime(time.time())
        elif (type(data) == IntType):
            data = time.localtime(data)
        elif (type(data) == FloatType):
            frac = data - int (data)
            data = time.localtime(int (data))

        self._data = time.strftime("%Y-%m-%dT%H:%M:%S", data)
        if frac:
            self._data += ("%g" % frac)[1:]
        self._data += 'Z'

    def __repr__(self):
        if self._name:
            return "<DateTime %s %s>" % (self._name, self._data)
        return "<DateTime %s>" % (self._data)

class Binary (NamedInstance):
    def __init__(self, data = None, name = None):
        self._name = name
        self._data = data
        if data:
            self._data = base64.encodestring(data)

class urType(NamedInstance):
    def __init__(self, data, name = None, typed = 1, datatype = ()):

        if self.__class__ == urType:
            raise Error, "urType can't be instantiated directly"

        NamedInstance.__init__(self, name)

        self._typename = self.__class__.__name__[:-4]
        self._cache = None

        if typed:
            self._ns = datatype[0]
            self._type = datatype[1] or self._typename
        else:
            self._ns, self._type = None, None

        self._data = self._checkValueSpace(data)

    def __str__(self):
        return str(self._data)

    def _checkValueSpace(self, data):
        return data

class stringType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD, None))

class booleanType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD, None))

    def __int__(self):
        return self._data

    __nonzero__ = __int__

    def __str__(self):
        return ['false', 'true'][self._data]

    def _checkValueSpace(self, data):
        if data in (0, 0.0, 'false', ''):
            return 0
        elif data in (1, 1.0, 'true'):
            return 1
        else:
            raise Error, "invalid %s value" % self._typename

class decimalType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD, None))

    def _checkValueSpace(self, data):
        if type(data) not in (IntType, LongType, FloatType):
            raise Error, "invalid %s value" % self._typename

        return data

class floatType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD, None))

    def _checkValueSpace(self, data):
        if type(data) not in (IntType, LongType, FloatType):
            raise Error, "invalid %s value" % self._typename

        return data

class doubleType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD, None))

    def _checkValueSpace(self, data):
        if type(data) not in (IntType, LongType, FloatType):
            raise Error, "invalid %s value" % self._typename

        return data

class durationType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD3, None))

    def _checkValueSpace(self, data):
        try:
            # A tuple or a scalar is OK, but make them into a list

            if type(data) == TupleType:
                data = list(data)
            elif type(data) != ListType:
                data = [data]

            if len(data) > 6:
                raise Exception, "too many values"

            # Now check the types of all the components, and find
            # the first nonzero element along the way.

            f = -1

            for i in range(len(data)):
                if data[i] == None:
                    data[i] = 0
                    continue

                if type(data[i]) not in \
                    (IntType, LongType, FloatType):
                    raise Exception, "element %d a bad type" % i

                if data[i] and f == -1:
                    f = i

            # If they're all 0, just use zero seconds.

            if f == -1:
                self._cache = 'P0S'

                return (0,) * 6

            # Make sure only the last nonzero element has a decimal
            # fraction and only the first element is negative.

            d = -1

            for i in range(f, len(data)):
                if data[i]:
                    if d != -1:
                        raise Exception, \
                            "all except the last nonzero element must be " \
                            "integers"
                    if data[i] < 0 and i > f:
                        raise Exception, \
                            "only the first nonzero element can be negative"
                    elif data[i] - long(data[i]) != 0:
                        d = i

            # Pad the list on the left if necessary.

            if len(data) < 6:
                n = 6 - len(data)
                f += n
                d += n
                data = [0] * n + data

            # Save index of the first nonzero element and the decimal
            # element for __str__.

            self.__firstnonzero = f
            self.__decimal = d

        except Exception, e:
            raise Error, "invalid %s value - %s" % (self._typename, e)

        return tuple(data)

    def __str__(self):
        if self._cache == None:
            d = self._data
            t = 0

            if d[self.__firstnonzero] < 0:
                s = '-P'
            else:
                s = 'P'

            t = 0

            for i in range(self.__firstnonzero, len(d)):
                if d[i]:
                    if i > 2 and not t:
                        s += 'T'
                        t = 1
                    if self.__decimal == i:
                        s += "%g" % abs(d[i])
                    else:
                        s += "%d" % abs(d[i])
                    s += ['Y', 'M', 'D', 'H', 'M', 'S'][i]

            self._cache = s

        return self._cache

class dateTimeType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD3, None))

    def _checkValueSpace(self, data):
        try:
            if data == None:
                data = time.time()

            if (type(data) == IntType):
                data = list(time.gmtime(data))
            elif (type(data) == FloatType):
                f = data - int(data)
                data = list(time.gmtime(int(data)))
                data[5] += f
            elif type(data) in (ListType, TupleType):
                if len(data) < 6:
                    raise Exception, "not enough values"
                if len(data) > 9:
                    raise Exception, "too many values"

                for i in range(len(data)):
                    if type(data[i]) not in \
                        (IntType, LongType, FloatType):
                        raise Exception, "element %d a bad type" % i

                if type(data) == TupleType:
                    data = list(data)

                if len(data) < 9:
                    data += [0] * (9 - len(data))
            else:
                raise Exception, "invalid type"

            data[8] = -1
        except Exception, e:
            raise Error, "invalid %s value - %s" % (self._typename, e)

        return tuple(data)

    def __str__(self):
        if self._cache == None:
            d = self._data
            s = ''

            s = time.strftime("%Y-%m-%dT%H:%M:%S", d)
            f = d[5] - int(d[5])
            if f != 0:
                s += ("%g" % f)[1:]
            s += 'Z'

            self._cache = s

        return self._cache

class timeType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD3, None))

    def _checkValueSpace(self, data):
        try:
            if data == None:
                data = time.gmtime(time.time())[3:6]
            elif type(data) in (IntType, FloatType):
                data = [data]

            if type(data) in (ListType, TupleType):
                if len(data) == 9:
                    data = data[3:6]
                elif len(data) > 3:
                    raise Exception, "too many values"

                for i in range(len(data)):
                    if type(data[i]) not in \
                        (IntType, LongType, FloatType):
                        raise Exception, "element %d a bad type" % i

                if type(data) == TupleType:
                    data = list(data)

                if len(data) < 3:
                    data = [0] * (3 - len(data)) + data

                for i in range(0, 2):
                    f = data[i] - int(data[i])
                    if f:
                        data[i + 1] += f * 60
                        data[i] = int(data[i])

                for i in range(2, 0, -1):
                    if data[i] > 60:
                        data[i - 1] += int(data[i] / 60)
                        data[i] %= 60

                if data[0] >= 24:
                    raise Exception, "too large"
            else:
                raise Exception, "invalid type"
        except Exception, e:
            raise Error, "invalid %s value - %s" % (self._typename, e)

        return tuple(data)

    def __str__(self):
        if self._cache == None:
            d = self._data
            s = ''

            s = time.strftime("%H:%M:%S", (0, 0, 0) + d + (0, 0, -1))
            f = d[2] - int(d[2])
            if f != 0:
                s += ("%g" % f)[1:]
            s += 'Z'

            self._cache = s

        return self._cache

class dateType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD3, None))

    def _checkValueSpace(self, data):
        try:
            if data == None:
                data = time.gmtime(time.time())[0:3]
            elif type(data) in (IntType, FloatType):
                data = time.gmtime(data)[0:3]

            if type(data) in (ListType, TupleType):
                if len(data) == 9:
                    data = data[0:3]
                elif len(data) < 3:
                    raise Exception, "too few values"
                elif len(data) > 3:
                    raise Exception, "too many values"

                for i in range(3):
                    if type(data[i]) not in \
                        (IntType, LongType, FloatType):
                        raise Exception, "element %d a bad type" % i

                if type(data) == ListType:
                    data = tuple(data)
            else:
                raise Exception, "invalid type"
        except Exception, e:
            raise Error, "invalid %s value - %s" % (self._typename, e)

        return data

    def __str__(self):
        if self._cache == None:
            self._cache = time.strftime("%Y-%m-%dZ",
                self._data + (0, 0, 0, 0, 0, -1))

        return self._cache


# gYearMonth
# gYear
# gMonthDay
# gDay
# gMonth

class hexBinaryType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD3, None))

    def __str__(self):
        if self._cache == None:
            self._cache = encodeHexString(self._data)

        return self._cache

class base64BinaryType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD3, None))

    def __str__(self):
        if self._cache == None:
            self._cache = base64.encodestring(self._data)

        return self._cache

class anyURIType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD, None))

# QName

class NOTATIONType(urType):
    def __init__(self, data, name = None, typed = 1, datatype = ()):

        if self.__class__ == NOTATIONType:
            raise Error, "a NOTATION can't be instantiated directly"

        urType.__init__(self, data, name, typed, datatype)

# normalizedString
# token
# language
# IDREFS
# ENTITIES
# NMTOKEN
# NMTOKENS
# Name
# NCName
# ID
# IDREF
# ENTITY

class integerType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD, None))

    def _checkValueSpace(self, data):
        if type(data) not in (IntType, LongType):
            raise Error, "invalid %s value" % self._typename

        return data

class nonPositiveIntegerType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD, None))

    def _checkValueSpace(self, data):
        if type(data) not in (IntType, LongType) or data > 0:
            raise Error, "invalid %s value" % self._typename

        return data

class negativeIntegerType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD, None))

    def _checkValueSpace(self, data):
        if type(data) not in (IntType, LongType) or data >= 0:
            raise Error, "invalid %s value" % self._typename

        return data

class longType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD, None))

    def _checkValueSpace(self, data):
        if type(data) not in (IntType, LongType):
            raise Error, "invalid %s value" % self._typename

        return data

class intType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD, None))

    def _checkValueSpace(self, data):
        if type(data) not in (IntType, LongType):
            raise Error, "invalid %s value" % self._typename

        return data

class shortType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD, None))

    def _checkValueSpace(self, data):
        if type(data) not in (IntType, LongType):
            raise Error, "invalid %s value" % self._typename

        return data

class byteType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD, None))

    def _checkValueSpace(self, data):
        if type(data) not in (IntType, LongType):
            raise Error, "invalid %s value" % self._typename

        return data

class nonNegativeIntegerType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD, None))

    def _checkValueSpace(self, data):
        if type(data) not in (IntType, LongType) or data < 0:
            raise Error, "invalid %s value" % self._typename

        return data

class unsignedLongType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD, None))

    def _checkValueSpace(self, data):
        if type(data) not in (IntType, LongType) or data < 0:
            raise Error, "invalid %s value" % self._typename

        return data

class unsignedIntType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD, None))

    def _checkValueSpace(self, data):
        if type(data) not in (IntType, LongType) or data < 0:
            raise Error, "invalid %s value" % self._typename

        return data

class unsignedShortType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD, None))

    def _checkValueSpace(self, data):
        if type(data) not in (IntType, LongType) or data < 0:
            raise Error, "invalid %s value" % self._typename

        return data

class unsignedByteType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD, None))

    def _checkValueSpace(self, data):
        if type(data) not in (IntType, LongType) or data < 0:
            raise Error, "invalid %s value" % self._typename

        return data

class positiveIntegerType(urType):
    def __init__(self, data, name = None, typed = 1):
        urType.__init__(self, data, name, typed, (NS.XSD, None))

    def _checkValueSpace(self, data):
        if type(data) not in (IntType, LongType) or data <= 0:
            raise Error, "invalid %s value" % self._typename

        return data

################################################################################
class RefHolder:
    def __init__(self, href):
        self.href = href

################################################################################
# Utility infielders
################################################################################
def collapseWhiteSpace(s):
    return re.sub(r'[\t\n\r ]+', ' ', s).strip()

def decodeHexString(self, data):
    high = {'0': 0x00, '1': 0x10, '2': 0x20, '3': 0x30, '4': 0x40,
        '5': 0x50, '6': 0x60, '7': 0x70, '8': 0x80, '9': 0x90, 'a': 0xa0,
        'b': 0xb0, 'c': 0xc0, 'd': 0xd0, 'e': 0xe0, 'f': 0xf0, 'A': 0xa0,
        'B': 0xb0, 'C': 0xc0, 'D': 0xd0, 'E': 0xe0, 'F': 0xf0,}
    low = {'0': 0x0, '1': 0x1, '2': 0x2, '3': 0x3, '4': 0x4,
        '5': 0x5, '6': 0x6, '7': 0x7, '8': 0x8, '9': 0x9, 'a': 0xa,
        'b': 0xb, 'c': 0xc, 'd': 0xd, 'e': 0xe, 'f': 0xf, 'A': 0xa,
        'B': 0xb, 'C': 0xc, 'D': 0xd, 'E': 0xe, 'F': 0xf,}

    bin = ''
    evenlen = len(data) & ~1

    for i in range(0, evenlen, 2):
        bin += chr(high[data[i]] + low[data[i + 1]])

    return bin

def encodeHexString(self, data):
    h = ''

    for i in data:
        h += "%02x" % ord(i)

    return h

class UnderflowError(exceptions.ArithmeticError):
    pass

################################################################################
# SOAP Parser
################################################################################
class SOAPParser(xml.sax.handler.ContentHandler):
    def __init__(self):
        xml.sax.handler.ContentHandler.__init__(self)
        self.body       = None
        self.header     = None
        self._data      = None
        self._next      = "ENV" # Keeping state for message validity
        self._type      = None
        self._name      = None
        self._attrs     = None
        # The stack stuff is reasonably complicated because it keeps track
        # of elements based on order and name.
        self._stack     = [] #[[]]  # start with an array for the body

        # Make two dictionaries to store the prefix <-> URI mappings, and
        # initialize them with the default
        self._prem      = {'xml': 'http://www.w3.org/XML/1998/namespace'}
        self._prem_r    = {'http://www.w3.org/XML/1998/namespace': 'xml'}
        self._href_map  = {}
        self._has_refs  = 0
        self._allrefs   = {}
        self._prules    = {}

    # Early attempt at adding automatic type conversion.
    def setParseRules(self, prules):
        self._prules = prules
        
    def startPrefixMapping(self, prefix, uri):
        self._prem[prefix] = uri
        self._prem_r[uri] = prefix

    def startElementNS(self, n, q, a):
        n = tuple(n) # Fix a 2.0 sax bug
        self._type = None
        self._data = ""
        self._id   = None

        href = a.get((None, "href"))
        if href:
            self._has_refs = 1
            self.addChild(n[1],RefHolder(href))
            self.getLast()._refmap_r[n[1]] = 1

        id = a.get((None,"id"))
        if id:
            self._has_refs = 1
            self._addId(n[1], id)

        if n == (NS.ENV, "Envelope") and self._next == "ENV":
            self._envns = a
            self._next = "HEADER"

        elif n == (NS.ENV, "Header") and self._next == "HEADER":
            self._next = "INHEADER"
            self.pushHeader() # for the header

        elif n == (NS.ENV, "Body") and self._next in ("BODY", "HEADER"):
            self._next = "ELEMS"
            self.pushBody() # for the Body

        elif self._next in ("INHEADER", "ELEMS"):
            self._name  = n
            self._attrs = a
            self._type = None
            for i in NS.XSI_L:
                self._type  = a.get((i,"type"))
                if self._type: break

            for i in a.getNames():
                if i[1] not in ("type", "null", "arrayType"):
                    if self.isStruct():
                        self.addAttr(n[1], tuple(i), str(a.get(i)))

            # Horrible hack to get around "untyped" arrays
            if self._type == None and a.get((NS.ENC, "arrayType")):
                try:
                    self._type = self._prem_r[NS.ENC] + ":Array"
                except:
                    self._type = None

            if self._type == None:
                if a.get((NS.XSI,"null")):
                    self._type = "null"
                elif self.isTArray(): # in a typed array
                    self._type = self.getTArray()
                    if not href:
                        if self._type[1] == "SOAPStruct":
                            self.pushStruct()
                    return
                elif n[1]=="Fault":
                    self.pushFault()
                else: # It's a struct
                    self.pushStruct()
                    # Ok if there are attributes we need to save them

            # type can still have a namespace, cause the parser won't fix it up.
            elif self._type.find(":")!=-1:
                (p1,p2) = self._type.split(":")
                p1 = self._prem[p1]
                self._type = [p1,p2]

                # Check for an array while we are here
                if self._type == [NS.ENC, "Array"]:
                    t = a.get((NS.ENC,"arrayType"))
                    if t[:8 + len(NS.XSD_T)] == NS.XSD_T+":ur-type":
                        self.pushArray()

                    else:
                        t = t.split("[")[0]
                        t = t.split(":")
                        t = [self._prem[t[0]],t[1]]
                        self.pushTArray(t)
                # XXX Hideous hack for microsoft's structType => ServerFault
                elif self._type[1] in ("SOAPStruct","ServerFault", "Struct",
                    "struct"):
                    # Another chance to get a struct
                    self.pushStruct()

        else:
            raise Error, "Badly formed SOAP message."

    def characters(self, c):
        if self._data != None:
            self._data += c

    def isArray(self):
        return self._stack and (self._stack[-1].__class__ in (Array, TypedArray))

    def isTArray(self):
        return self._stack and (self._stack[-1].__class__ == TypedArray)

    def isStruct(self):
        return self._stack and issubclass(self._stack[-1].__class__, Struct)

    def pushArray(self, name=None):
        self._stack.append(Array(name))

    def pushTArray(self, tarray, name=None):
        self._stack.append(TypedArray(name, tarray))

    def pushStruct(self, name=None):
        self._stack.append(Struct(name))

    def pushBody(self):
        self._stack.append(Body())

    def pushHeader(self):
        self._stack.append(Header())

    def pushFault(self):
        self._stack.append(Fault())

    def getLast(self):
        return self._stack[-1]

    def _addId(self, name, id):
        t = self.getLast()
        if t._refmap.has_key(name):
            t._refmap[name].append(id)
        else:
            t._refmap[name] = [id]

    def addChild(self, name, value):
        t = self.getLast()
        # Before we add see if it's a reference
        if t._refmap_r.has_key(name):
            del t._refmap_r[name]
            return

        if self._has_refs:
            if name in t._refmap.keys():
                t._refmap_r[t._refmap[name][0]] = value
                self._allrefs[t._refmap[name][0]]= value
                t._refmap[name] = t._refmap[name][1:]
                #return

        if self.isArray():
            t.data.append(value)
        else:
            t.__dict__[name] = value
            t._asdict[name]  = value  # for convenience
            t._keyord.append(name)
            t._aslist += (value,)

    def addAttr(self, tag, name, value):
        t = self.getLast()
        t._setAttr(tag, name, value)
        t._setAttr(len(t._keyord), name, value)

    def getTArray(self):
        return self._stack[-1]._type

    def stackPop(self):
        self._stack.pop()

    def endElementNS(self, n, q):
        n = tuple(n) # Fix a 2.0 sax bug
        if n == (NS.ENV, "Envelope"): return
        try:
            d = self._data.strip()
        except:
            d = ""

        try:
            array = self._type[1]=='Array'
        except:
            array = 0

        if ((not self._type) and (not d)) or self._type == "null" or array:
            if self._type == "null":
                t = None
            elif self.isArray():  # Array
                t = self.getLast()
                t._name = n[1]

                self.stackPop()

            elif n[1] == "Fault":
                t = self.getLast()
                self.stackPop()

            # Nothing, some parsers expect this to be a null
            # string so let's try it
            # XXX Not fool proof, as you could have a null
            # string down in a struct with a namespace
            elif len(self.getLast()._aslist) == 0 and not n[0]:
                self.stackPop()
                t = ''
            else:
                add_to_parent = 1
                if n[1]=="Header":
                    t = self.header = self.getLast()
                    add_to_parent = 0
                    self._next = "BODY"
                elif n[1]=="Body":
                    t = self.body = self.getLast()
                    add_to_parent = 0
                else:
                    t = self.getLast()

                t._name = n[1] # Set name
                t._ns = n[0]   # Set namespace
                self.stackPop()
                if not add_to_parent: return

            # put the data up one in the stack.
            self.addChild(n[1], t)

        else:
            # this is terrible but, if there is data, and no type
            # it's an untyped data, not a struct so pop it

            if (self._type==None) and (not self.isTArray()):
                self.stackPop()

            d = self.convertType()

            if d != None:
                self.addChild(n[1], d)

        # Clear the data
        self._data = None
        self._name = None
        self._type = None
        self._id   = None

    def fixup(self, start, prev, prules={}):
        if prules.has_key(start._name):
            prules = prules[start._name]

        if issubclass(start.__class__,Struct):
            count = 0
            for i in start._aslist:
                try:
                    if i.__class__ == RefHolder:
                        id = i.href.split("#")[1]
                        #try:
                        #    tv = prev._refmap_r[id]
                        #except:
                        tv = self._allrefs[id]
                        start._aslist[count] = tv
                        start.__dict__[start._keyord[count]] = tv

                except:
                    pass
                try:
                    if i.__dict__ or i.data:
                        self.fixup(i, start, prules)
                except:
                    pass
                count += 1
            
            # Now fix up types
            if prules:
                for (k,v) in start.__dict__.items():
                    if prules.has_key(k):
                        # Right now strings and funcs, maybe classes someday
                        if type(prules[k]) == StringType:
                            
                            if prules[k].find("arrayType") != -1:
                                ty = prules[k].split("=")[1]
                                ty = ty.split("[")[0]
                                typed = (ty != "ur-type")
                                arr = []
                                for i in v._aslist:
                                    if not typed:
                                        arr.append(i)
                                    else:
                                        arr.append(self.convertType(i, (NS.XSD, ty)))
                                if typed:
                                    arr = TypedArray(k, (NS.XSD, ty),arr)
                                else:
                                    arr = Array(k, arr)
                                start.__dict__[k] = arr
                                # this is ugly but find it in order as well
                                count = 0
                                for i in start._aslist:
                                    if i==v:
                                        break
                                    count += 1
                                start._aslist[count] = arr
                                        
                            else:
                                start.__dict__[k] = self.convertType(v,(NS.XSD, prules[k]))
                        elif callable(prules[k]):
                            start.__dict__[k] = prules[k](v)
                        elif type(prules[k]) == DictType:
                            # Eventually check for structure
                            pass
                        
        elif issubclass(start.__class__,Array):
            count = 0
            for i in start.data:
                try:
                    if i.__class__ == RefHolder:
                        id = i.href.split("#")[1]
                        #start.data[count] = prev._refmap_r[id]
                        start.data[count] = self._allrefs[id]
                except:
                    pass
                try:
                    if i.__dict__ or i.data:
                        self.fixup(v, start, prules)
                except:
                    pass
                count += 1

    def endDocument(self):
        if self._has_refs or self._prules:
            self.fixup(self.body, self.body, self._prules)

    # Convertion
    class DATETIMERE:
        YEAR = '(?P<year>-?\d{4,})'
        MONTH = '(?P<month>\d{2})'
        DAY = '(?P<day>\d{2})'
        HOUR = '(?P<hour>\d{2})'
        MINUTE = '(?P<minute>\d{2})'
        SECOND = '(?P<second>\d{2}(?:\.\d*)?)'
        TIMEZONE = '(?P<zulu>Z)|(?P<tzsign>[-+])(?P<tzhour>\d{2}):' \
            '(?P<tzminute>\d{2})'

        all = {'year': YEAR, 'month': MONTH, 'day': DAY, 'hour': HOUR,
            'minute': MINUTE, 'second': SECOND, 'timezone': TIMEZONE}

        dateTime = '^%(year)s-%(month)s-%(day)sT' \
            '%(hour)s:%(minute)s:%(second)s(%(timezone)s)?$' % all
        timeInstant = dateTime
        time = '^%(hour)s:%(minute)s:%(second)s(%(timezone)s)?$' % all
        date = '^%(year)s-%(month)s-%(day)s(%(timezone)s)?$' % all
        gYearMonth = '^%(year)s-%(month)s(%(timezone)s)?$' % all
        gYear = '^%(year)s(%(timezone)s)?$' % all
        gMonthDay = '^--%(month)s-%(day)s(%(timezone)s)?$' % all
        gDay = '^---%(day)s(%(timezone)s)?$' % all
        gMonth = '^--%(month)s--(%(timezone)s)?$' % all

    def convertDateTime(self, value, kind):
        if not hasattr(self.DATETIMERE, kind):
            return None

        exp = getattr(self.DATETIMERE, kind)

        if type(exp) == type(''):
            exp = re.compile(exp)
            setattr (self.DATETIMERE, kind, exp)

        m = exp.search(value)

        if m == None:
            return value

        zoffs = 0

        try:
            if m.group('zulu') == None:
                zoffs = 60 * (60 * int(m.group('tzhour')) +
                    int(m.group('tzminute')))
            if m.group('tzsign') == '-':
                zoffs = -zoffs
        except:
            pass

        d = {'year': 1970, 'month': 1, 'day': 1,
            'hour': 0, 'minute': 0, 'second': 0}

        for i in d.keys():
            try:
                s = m.group(i)
                if s != None:
                    d[i] = float(s)
            except:
                pass

        # Why the int around time.mktime? time.mktime seems to ignore any
        # fractional seconds, so we add the fractional seconds back in.
        # However, if time.mktime is ever changed to not ignore the
        # fractional seconds, taking int(time.mktime) assures this won't
        # break.

        return int(time.mktime((d['year'], d['month'], d['day'], d['hour'],
            d['minute'], d['second'], 0, 0, -1))) + \
                (d['second'] - int(d['second'])) - zoffs

    intlimits = \
    {
        'nonPositiveInteger':   (None, 0),
        'negativeInteger':      (None, -1),
        'long':                 (-9223372036854775808L, 9223372036854775807L),
        'int':                  (-2147483648L, 2147483647),
        'short':                (-32768, 32767),
        'byte':                 (-128, 127),
        'nonNegativeInteger':   (0, None),
        'positiveInteger':      (1, None),
        'unsignedLong':         (0, 18446744073709551615L),
        'unsignedInt':          (0, 4294967295L),
        'unsignedShort':        (0, 65535),
        'unsignedByte':         (0, 255),
    }
    floatlimits = \
    {
        'float':        (7.0064923216240861E-46, -3.4028234663852886E+38,
                         3.4028234663852886E+38),
        'double':       (2.4703282292062327E-324, -1.7976931348623158E+308,
                         1.7976931348623157E+308),
    }
    zerofloatre = '[1-9]'

    def convertType(self, data=None, datatype=None):
        d = data or self._data
        t = datatype or self._type
        if not t: return d # XXX how do we get in here with a None?
        if t[0] in NS.EXSD_L:
            if t[1] == "integer":
                try: return int(d)
                except: return long(d)
            if self.intlimits.has_key (t[1]):
                l = self.intlimits[t[1]]
                try: d = int(d)
                except: d = long(d)

                if l[0] != None and d < l[0]:
                    raise UnderflowError, "%s too small"
                if l[1] != None and d > l[1]:
                    raise OverflowError, "%s too large"

                return d
            if t[1] in ("string", "QName", "NOTATION", "NMTOKEN", "Name",
                "NCName", "ID", "IDREF", "Entity"):
                try:
                    return str(d)
                except:
                    return d
            if t[1] == "boolean":
                try:
                    return int(d)   # Better as int than class
                except:
                    s = d.strip().lower()
                    if s == 'false':
                        return 0
                    if s == 'true':
                        return 1
                    raise AttributeError, "invalid boolean value"
            if self.floatlimits.has_key (t[1]):
                l = self.floatlimits[t[1]]
                s = d.strip().lower()
                try:
                    d = float(s)
                except:
                    # Some platforms don't implement the float stuff. This
                    # is close, but NaN won't be > "INF" as required by the
                    # standard.

                    if s in ("nan", "inf"):
                        return 1e300**2
                    if s == "-inf":
                        return -1e300**2

                    raise

                if str (d) == 'nan':
                    if s != 'nan':
                        raise ValueError, "invalid %s" % t[1]
                elif str (d) == '-inf':
                    if s != '-inf':
                        raise UnderflowError, "%s too small" % t[1]
                elif str (d) == 'inf':
                    if s != 'inf':
                        raise OverflowError, "%s too large" % t[1]
                elif d < 0:
                    if d < l[1]:
                        raise UnderflowError, "%s too small" % t[1]
                elif d > 0:
                    if d < l[0] or d > l[2]:
                        raise OverflowError, "%s too large" % t[1]
                elif d == 0:
                    if type(self.zerofloatre) == StringType:
                        self.zerofloatre = re.compile(self.zerofloatre)

                    if self.zerofloatre.search(s):
                        raise UnderflowError, "invalid %s" % t[1]

                return d
            if t[1] in ("dateTime", "timeInstant"):
                return self.convertDateTime(d, t[1])
            if t[1] == "decimal":
                try:
                    return float(d)
                except:
                    return d
            if t[1] in ("language", "IDREFS", "ENTITIES", "NMTOKENS"):
                return collapseWhiteSpace(d)
        if t[0] in NS.XSD_L:
            if t[1] in ("base64", "base64Binary"):
                return base64.decodestring(d)
            if t[1] == "hexBinary":
                return decodeHexString(d)
            if t[1] == "anyURI":
                return urllib.unquote(d)
            if t[1] == "normalizedString":
                return str(d)
            if t[1] in ("token"):
                return collapseWhiteSpace(d)
        if t[0] == NS.ENC:
            if t[1] == "base64":
                return base64.decodestring(d)
            if t[1] == "uriReference":
                return urllib.unquote(d)
        return d # No type

################################################################################
# call to SOAPParser that keeps all of the info
################################################################################
def _parseSOAP(xml_str, parse_rules={}):
    try:
        from cStringIO import StringIO
    except ImportError:
        from StringIO import StringIO

    parser = xml.sax.make_parser()
    t = SOAPParser()
    if parse_rules:
        t.setParseRules(parse_rules)
    parser.setContentHandler(t)
    e = xml.sax.handler.ErrorHandler()
    parser.setErrorHandler(e)

    inpsrc = xml.sax.xmlreader.InputSource()
    inpsrc.setByteStream(StringIO(xml_str))

    # turn on namespace mangeling
    parser.setFeature(xml.sax.handler.feature_namespaces,1)

    parser.parse(inpsrc)

    return t

################################################################################
# SOAPParser's more public interface
################################################################################
def parseSOAP(xml_str):
    t = _parseSOAP(xml_str)
    return t.body


def parseSOAPRPC(xml_str,header=0, parse_rules={}):
    t = _parseSOAP(xml_str, parse_rules=parse_rules)
    p = t.body._aslist[0]

    # Empty string, for RPC this translates into a void
    if type(p) in (type(''), type(u'')) and p.strip() in ('', u''):
        name = "Response"
        for k in t.body.__dict__.keys():
            if k[0]!="_":
                name = k
        p = Struct(name )

    if header:
        return (p, t.header)
    else:
        return p


################################################################################
# SOAP Builder
################################################################################
class SOAPBuilder:
    _xml_top = '<?xml version="1.0" encoding="%s"?>\n'
    _env_top = '%(ENV_T)s:Envelope %(ENV_T)s:encodingStyle="%(ENC)s"' % \
        NS.__dict__
    _env_bot = '</%(ENV_T)s:Envelope>\n' % NS.__dict__

    # Namespaces potentially defined in the Envelope tag.

    _env_ns = {NS.ENC: NS.ENC_T, NS.ENV: NS.ENV_T,
        NS.XSD: NS.XSD_T, NS.XSD3: NS.XSD3_T, NS.XSI: NS.XSI_T}

    def __init__(self, args = (), kw = {}, method = None, namespace = None,
        header = None, methodattrs = None, envelope = 1, encoding = 'UTF-8',
                 use_refs=0):

        # Test the encoding, raising an exception if it's not known
        ''.encode(encoding)

        if type(args) != TupleType:
            self.args  = (args,)
        else:
            self.args  = args

        self.kw         = kw
        self.envelope   = envelope
        self.body       = 1
        self.encoding   = encoding
        self.method     = method
        self.namespace  = namespace
        self.header     = header
        self.methodattrs= methodattrs
        self.use_refs   = use_refs
        self.out        = ''
        self.tcounter   = 0
        self.ncounter   = 0
        self.icounter   = 1
        self.envns      = {}
        self.ids        = {}
        self.ids_r      = {}
        self._ordered_m = 0
        self._current   = 0
        self._ordattrs  = None

        # If we are passed in a body we'll use theirs
        try:
            if args.__class__ == Body:
                self.body  = 0
        except:
            pass

    def build(self):
        ns_map = {}

        if self.header:
            # Create a header.
            # Call genns to record that we've used SOAP-ENV.
            ns = self.genns(ns_map, NS.ENV)[0]
            self.dump(self.header, "%s:Header" % ns[0], ns_map = ns_map)
            self.header = None # Wipe it out so no one is using it.
        if self.body:
            # Call genns to record that we've used SOAP-ENV.
            body_ns = self.genns(ns_map, NS.ENV)[0]
            self.out += "<%s:Body>\n" % body_ns[0]
        if self.method:
            a = ''
            if self.methodattrs:
                for (k,v) in self.methodattrs.items():
                    a += ' %s="%s"' % (k,v)

            n = ''
            if self.namespace:  # Use the namespace info handed to us
                ns, n = self.gennsattr(ns_map, self.namespace)
                self.method = ns[0] + ":" + self.method

            self.out += '<%s%s%s>\n' % (self.method, n, a)

        try:
            for i in self.args:
                self.dump(i, ns_map = ns_map)
                
            for (k,v) in self.kw.items():
                self.dump(v, k, ns_map = ns_map)
        except Error, e:
            if (e.msg=="Cannot serialize recursive object") \
               and self.use_refs==0:
                # restart
                b = SOAPBuilder(args=self.args,kw=self.kw,method=self.method, namespace=self.namespace, header=self.header, methodattrs=self.methodattrs, envelope=self.envelope, encoding=self.encoding, use_refs=1)
                return b.build()

            else:
                raise e
                
            

        if self.method:
            self.out += "</%s>\n" % self.method
        if self.body:
            self.out += "</%s:Body>\n" % body_ns[0]

        if self.envelope:
            e = map (lambda ns: 'xmlns:%s="%s"' % (ns[1], ns[0]),
                self.envns.items())

            self.out = '<' + self._env_top + ' '.join([''] + e) + '>\n' + \
                self.out + \
                self._env_bot

        self.out = self._xml_top % self.encoding + self.out

        return self.out.encode(self.encoding)

    def gentag(self):
        self.tcounter += 1
        return "v%d" % self.tcounter

    def genns(self, ns_map, nsURI):
        if type(nsURI) == TupleType: # already a tuple
            if len(nsURI)==2:
                ns, nsURI = nsURI
            else:
                ns, nsURI = None, nsURI[0]
        else:
            ns = None
        if ns_map.has_key(nsURI):
            return ((ns_map[nsURI], nsURI), 0)
        if self._env_ns.has_key(nsURI):
            ns = self.envns[nsURI] = ns_map[nsURI] = self._env_ns[nsURI]
            return ((ns, nsURI), 0)
        if not ns:
            ns = "ns%d" % self.ncounter
            self.ncounter += 1
        ns_map[nsURI] = ns
        return ((ns, nsURI), 1)

    def gennsattr(self, ns_map, nsURI):
        ns, newns = self.genns(ns_map, nsURI)
        if newns:
            return (ns, ' xmlns:%s="%s"' % ns)
        return (ns, '')

    def genid(self):
        if self.use_refs:
            id = ' id="%s"' % self.icounter
            self.icounter += 1
            return id
        else:
            return ''

    def check_id(self, obj):
        if self.ids.has_key(id(obj)):
            if self.use_refs==0:
                raise Error, "Cannot serialize recursive object"
            
            return self.ids[id(obj)]
            pass
        else:  
            self.ids[id(obj)] = self.icounter+1
            #self.ids_r[self.icounter] = obj
            #self.icounter += 1
            return None

    def build_attr(self, ns_map, attr_map, attr_ord, tag):
        a = ''
        if attr_ord != ():
            try:
                for k, v in attr_ord[1][attr_ord[0]].items():
                    ns, n = self.gennsattr(ns_map, k[0])
                    a += n + ' %s:%s="%s"' % (ns[0], k[1], v)
            except:
                pass
        else:
            if attr_map:
                try:
                    for k, v in attr_map[tag].items():
                        ns, n = self.gennsattr(ns_map, k[0])
                        a += n + ' %s:%s="%s"' % (ns[0], k[1], v)
                except:
                    pass
        return a

    # dumpers

    def dump(self, obj, tag = None, typed = 1, ns_map = {},
        attr_map = {}, attr_ord = ()):

        ns_map = ns_map.copy()
        if self.header:
            try:
                if obj.__class__ == Header:
                    attr_map.update(obj._attr_map)
            except:
                pass
        else:
            try:
                attr_map.update(obj._attr_map)
            except:
                pass
        try:
            meth = getattr(self, "dump_" + type(obj).__name__)
            meth(obj, tag, typed, ns_map, attr_map, attr_ord)
        except AttributeError:
            # Use the generic dumper, right now probably only for int
            self.out += self.dumper(NS.XSD, type(obj).__name__, \
                obj, tag, typed, ns_map, attr_map, attr_ord)

    # generic dumper
    def dumper(self, nsURI, obj_type, obj, tag, typed = 1, ns_map = {},
        attr_map = None, attr_ord = (),
        xml = '<%(tag)s%(type)s%(ns)s%(id)s%(attrs)s>%(obj)s</%(tag)s>\n'):

        tag = tag or self.gentag()

            
        t = n = ''
        if typed and obj_type:
            ns, n = self.gennsattr(ns_map, nsURI)
            ins = self.genns(ns_map, NS.XSI)[0]
            t = ' %s:type="%s:%s"' % (ins[0], ns[0], obj_type)

        return xml % {"tag": tag, "type": t, "obj": obj, "ns": n,
                      "id" : self.genid(), "attrs": \
                      self.build_attr(ns_map, attr_map, attr_ord,tag)}

    def dump_float(self, obj, tag, typed = 1, ns_map = {},
        attr_map = None, attr_ord = ()):

        # Terrible windows hack
        if not good_float:
            if obj == float(1e300**2):
                obj = "INF"
            elif obj == float(-1e300**2):
                obj = "-INF"

        obj = str(obj)
        if obj in ('inf', '-inf'):
            obj = str(obj).upper()
        elif obj == 'nan':
            obj = 'NaN'
        self.out += self.dumper(NS.XSD, "float", obj, tag, typed, ns_map,
            attr_map, attr_ord)

    def dump_string(self, obj, tag, typed = 0, ns_map = {},
        attr_map = None, attr_ord = ()):

        self.out += self.dumper(NS.XSD, "string", saxutils.escape(obj),
            tag, typed, ns_map, attr_map, attr_ord)

    dump_unicode = dump_string

    def dump_None(self, obj, tag, typed = 0, ns_map = {},
        attr_map = None, attr_ord = ()):

        tag = tag or self.gentag()
        ns = self.genns(ns_map, NS.XSI)[0]
        self.out += '<%s %s:null="1"%s/>\n' % (tag, ns[0],
            self.build_attr(ns_map, attr_map, attr_ord, tag))

    def dump_list(self, obj, tag, typed = 1, ns_map = {},
        attr_map = None, attr_ord = ()):

        self.check_id(obj)
        tag = tag or self.gentag()
        same_type = 1
        empty = 0
        try:
            last = obj[0]
        except:
            empty = 1
            same_type = 1
            last = Struct()

        # First scan list to see if all are the same type
        if not empty:
            for i in obj[1:]:
                if type(last) != type(i) or \
                    (type(last) == type(self) and \
                        last.__class__ != i.__class__):
                    same_type = 0
                    break
                last = i

        t = self.genns(ns_map, NS.XSD)[0][0] + ":ur-type"

        n = ''
        if same_type:
            if (type(last) == type(self) and isinstance(last, Struct)) or \
                type(last) == DictType: # force to urn struct
                ns, n = self.gennsattr(ns_map, NS.URN)
                t = ns[0] + ":SOAPStruct"
            elif isinstance(last, urType):
                if last._ns:
                    ns, n = self.gennsattr(ns_map, last._ns)
                    t = ns[0] + ":" + last._type
                else:
                    t = 'ur-type'
            else:
                t = self.genns(ns_map, NS.XSD)[0][0] + ":" + \
                    type(last).__name__

        a = self.build_attr(ns_map, attr_map, attr_ord,tag)

        if n == '':
            ens, n = self.gennsattr(ns_map, NS.ENC)
        else:
            ens = self.genns(ns_map, NS.ENC)[0]
        ins = self.genns(ns_map, NS.XSI)[0]
        self.out += \
            '<%s %s:arrayType="%s[%d]" %s:type="%s:Array"%s%s%s>\n' %\
            (tag, ens[0], t, len(obj), ins[0], ens[0], n, self.genid() ,a)

        typed = not same_type

        for i in obj:
            self.dump(i, "item", typed, ns_map, attr_map, attr_ord)

        self.out += '</%s>\n' % tag

    dump_tuple = dump_list

    def dump_dictionary(self, obj, tag, typed = 1, ns_map = {},
        attr_map = None, attr_ord = ()):

        self.check_id(obj)
        tag = tag or self.gentag()

        self.out += '<%s%s%s>\n' % (tag, self.genid(),
            self.build_attr(ns_map, attr_map, attr_ord, tag))

        for (k,v) in obj.items():
            if k[0]!="_":
                self.dump(v, k, 1, ns_map, attr_map, attr_ord)

        self.out += '</%s>\n' % tag

    def dump_instance(self, obj, tag, typed = 1, ns_map = {},
        attr_map = None, attr_ord = ()):

        id = self.check_id(obj)
        if id:
            self.out += '<%s href="#%d"/>\n' % (tag, id-1)
            return

        a = self.build_attr(ns_map, attr_map, attr_ord, tag)

        if not tag:
            # If it has a name use it.
            if issubclass(obj.__class__, NamedInstance) and obj._name:
                tag = obj._name
            else:
                tag = self.gentag()

        t = ''

        if isinstance(obj, urType):
            if typed and obj._ns:
                ns, n = self.gennsattr(ns_map, obj._ns)
                ins = self.genns(ns_map, NS.XSI)[0]
                t = ' %s:type="%s:%s"%s' % (ins[0], ns[0], obj._type, n)

            self.out += '<%s%s%s%s>%s</%s>\n' % (tag, t, self.genid(), a, obj, tag)

        elif isinstance(obj, Array):      # Array
            self.dump_list(obj.data, tag, typed, ns_map, attr_map, attr_ord)

        elif isinstance(obj, Void):     # void
            self.out += "<%s></%s>\n" % (tag,tag)

        elif isinstance(obj, Fault):    # Fault
            ns, n = self.gennsattr(ns_map, NS.ENV)
            self.out += """<%s:Fault%s>
<faultcode>%s</faultcode>
<faultstring>%s</faultstring>
""" % (ns[0], n, obj.faultcode, obj.faultstring)
            if hasattr(obj, "detail"):
                self.dump(obj.detail, "detail", typed, ns_map)
            self.out += "</%s:Fault>\n" % ns[0]

        elif isinstance(obj, Boolean):  # Boolean
            if typed:
                ns, n = self.gennsattr(ns_map, NS.XSD)
                ins = self.genns(ns_map, NS.XSI)[0]
                t = ' %s:type="%s:boolean"%s' % (ins[0], ns[0], n)
            self.out += '<%s%s%s%s>%d</%s>\n' % (tag, t, self.genid(),a, obj, tag)

        elif isinstance(obj, DateTime): # DateTime
            if typed:
                ns, n = self.gennsattr(ns_map, NS.XSD)
                ins = self.genns(ns_map, NS.XSI)[0]
                t = ' %s:type="%s:timeInstant"%s' % (ins[0], ns[0], n)
            self.out += '<%s%s%s%s>%s</%s>\n' % (tag, t, self.genid(),a, obj._data, tag)

        elif isinstance(obj, Binary):   # Binary
            if typed:
                ns, n = self.gennsattr(ns_map, NS.ENC)
                ins = self.genns(ns_map, NS.XSI)[0]
                t = ' %s:type="%s:base64"%s' % (ins[0], ns[0], n)
            self.out += '<%s%s%s%s>%s</%s>\n' % (tag, t, self.genid(),a, obj._data, tag)

        elif obj.__class__ in (Struct, Body, Header):   # Struct
            # Check for namespace
            n = ''
            if obj._ns:
                ns, n = self.gennsattr(ns_map, obj._ns)
                tag = ns[0] + ":" + tag
            self.out += "<%s%s%s%s>\n" % (tag, n, self.genid(),a)

            # If we have order use it.
            order = 1

            for i in obj._keys():
                if i not in obj._keyord:
                    order = 0
            if order:
                count = 0
                for i in range(len(obj._keyord)):
                    if obj._attr_ord == {}:
                        t_a_o = ()
                    else:
                        t_a_o = (count, obj._attr_ord)

                    self.dump(obj._aslist[i], obj._keyord[i], 1, ns_map,
                        attr_map, t_a_o)
                    count += 1
                self.out += '</%s>\n' % tag
                return
            # don't have pristine order information, just build it.
            for (k,v) in obj.__dict__.items():
                if k[0]!="_":
                    self.dump(v, k, 1, ns_map, attr_map, attr_ord)

            self.out += '</%s>\n' % tag

        else:                           # Some Class
            self.out += '<%s%s%s>\n' % (tag, self.genid(),a)
            for (k,v) in obj.__dict__.items():
                if k[0] != "_":
                    self.dump(v, k, 1, ns_map, attr_map, attr_ord)

            self.out += '</%s>\n' % tag


################################################################################
# SOAPBuilder's more public interface
################################################################################
def buildSOAP(args=(), kw={}, method=None, namespace=None, header=None,
              methodattrs=None,envelope=1,encoding='UTF-8'):
    t = SOAPBuilder(args=args,kw=kw, method=method, namespace=namespace,
        header=header, methodattrs=methodattrs,envelope=envelope,
        encoding=encoding)
    return t.build()

################################################################################
# RPC
################################################################################

def SOAPUserAgent():
    return "SOAP.py " + __version__ + " (actzero.com)"

################################################################################
# Client
################################################################################
class HTTPTransport:
    # Need a Timeout someday?
    def call(self, proxy, data, soapaction = None, encoding = None):
        import httplib
        import urllib

        type, uri = urllib.splittype(proxy)
        host, path = urllib.splithost(uri)

        # Build a request
        r = httplib.HTTP(host)
        r.putrequest("POST", path)

        r.putheader("Host", host)
        r.putheader("User-agent", SOAPUserAgent())
        s = 'text/xml';
        if encoding != None:
            s += '; charset="%s"' % encoding
        r.putheader("Content-type", s)
        r.putheader("Content-length", str(len(data)))

        if DEBUG:
            print "***HTTP_HEADER***"
            print "POST %s HTTP/1.0" % path
            print "Host:", host
            print "User-agent:", "SOAP.py " + __version__ + " (actzero.com)"
            print "Content-type:", s
            print "Content-length:", str(len(data))

        if not soapaction:
            r.putheader("SOAPAction", '""')
            if DEBUG: print "SOAPAction:", '""'
        else:
            r.putheader("SOAPAction", '"%s"' % soapaction)
            if DEBUG: print "SOAPAction:", '"%s"' % soapaction
        if DEBUG: print "*****************"

        r.endheaders()

        if DEBUG : print "\n***SENDING***\n", data, "*" * 13 + "\n"

        # send the payload
        r.send(data)

        # read response line
        code, msg, headers = r.getreply()

        if code not in (200, 500):
            raise HTTPError(code, msg)

        # return response payload
        return r.getfile().read()

################################################################################
# SOAP Proxy
################################################################################
class SOAPProxy:
    def __init__(self, proxy, namespace = None, soapaction = None,
        header = None, methodattrs = None, transport = HTTPTransport,
        encoding = 'UTF-8', throw_faults = 1, unwrap_results = 1):

        # Test the encoding, raising an exception if it's not known
        ''.encode(encoding)

        self.proxy          = proxy
        self.namespace      = namespace
        self.soapaction     = soapaction
        self.header         = header
        self.methodattrs    = methodattrs
        self.transport      = transport()
        self.encoding       = encoding
        self.throw_faults   = throw_faults
        self.unwrap_results = unwrap_results

    def __call(self, name, args, kw, ns=None, sa=None, hd=None, ma=None):
        ns = ns or self.namespace
        ma = ma or self.methodattrs

        if sa: # Get soapaction
            if type(sa) == TupleType: sa = sa[0]
        else:
            sa = self.soapaction

        if hd: # Get header
            if type(hd) == TupleType: hd = hd[0]
        else:
            hd = self.header

        hd = hd or self.header

        if ma: # Get methodattrs
            if type(ma) == TupleType: ma = ma[0]
        else:
            ma = self.methodattrs
        ma = ma or self.methodattrs

        m = buildSOAP(args=args, kw=kw, method=name, namespace=ns, header=hd,\
                      methodattrs=ma, encoding = self.encoding)

        r = self.transport.call(self.proxy, m, sa, encoding = self.encoding)
        if DEBUG :
            print "\n***RECEIVING***\n", r, "*" * 13 + "\n"
            sys.stdout.flush()
        p = parseSOAPRPC(r)

        try:
            throw_struct = self.throw_faults and p._name == "Fault"
        except:
            throw_struct = 0

        if throw_struct:
            raise p

        # Bubble a regular result up, if there is only element in the
        # struct, assume that is the result and return it.
        # Otherwise it will return the struct with all the elements
        # as attributes.
        if self.unwrap_results:
            try:
                count = 0
                for i in p.__dict__.keys():
                    if i[0] != "_":  # don't move the private stuff
                        count += 1
                        t = getattr(p, i)
                if count == 1: p = t # Only one piece of data, bubble it up
            except:
                pass
        return p

    def _callWithBody(self, body):
        return self.__call(None, body, {})

    def __getattr__(self, name):  # hook to catch method calls
        return self.__Method(self.__call, name)

    # To handle attribute wierdness
    class __Method:
    # some magic to bind an SOAP method to an RPC server.
    # supports "nested" methods (e.g. examples.getStateName)
    # -- Concept borrowed from xmlrpc/soaplib -- www.pythonware.com
    # Altered (improved?) to let you inline namespaces on a per
    # call basis ala SOAP::LITE -- www.soaplite.com
        def __init__(self, call, name, ns = None, sa = None, hd = None, ma = None):
            self.__call = call
            self.__name = name
            self.__ns   = ns
            self.__sa   = sa
            self.__hd   = hd
            self.__ma   = ma
            if self.__name[0] == "_":
                self.__call__ = self.__f_call
            else:
                self.__call__ = self.__r_call

        def __getattr__(self, name):
            if self.__name[0] == "_":
                # Don't nest method if it is a directive
                return self.__class__(self.__call, "%s" % name, self.__ns, \
                                      self.__sa, self.__hd, self.__ma)

            return self.__class__(self.__call, "%s.%s" % (self.__name, name), \
                                  self.__ns, self.__sa, self.__hd, self.__ma)

        def __f_call(self, *args, **kw):
            if self.__name == "_ns" : self.__ns = args
            elif self.__name == "_sa" : self.__sa = args
            elif self.__name == "_hd" : self.__hd = args
            elif self.__name == "_ma" : self.__ma = args
            return self

        def __r_call(self, *args, **kw):
            return self.__call(self.__name, args, kw, self.__ns, self.__sa, \
                               self.__hd, self.__ma)

################################################################################
# Server
################################################################################

# Method Signature class for adding extra info to registered funcs, right now
# used just to indicate it should be called with keywords, instead of ordered
# params.
class MethodSig:
    def __init__(self, func, keywords=1):
        self.func     = func
        self.keywords = keywords
        self.__name__ = func.__name__

    def __call__(self, *args, **kw):
        return apply(self.func,args,kw)

# A class to describe how header messages are handled
class HeaderHandler:
    # Initially fail out if there are any problems.
    def __init__(self, header):
        for i in header.__dict__.keys():
            fault = 0
            try:
                fault = int(header._getMustUnderstand(i))
            except:
                pass
            if fault:
                raise Fault, ("%s:MustUnderstand" % NS.ENV_T,
                    "Can't process the request because this "
                    "implementation does not support the mustUnderstand "
                    "attribute.")


################################################################################
# SOAP Server
################################################################################
class SOAPServer(SocketServer.TCPServer):
    import BaseHTTPServer
    class SOAPRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
        def version_string(self):
            return '<a href="http://www.actzero.com/solution.html">' + \
                'SOAP.py ' + __version__ + '</a> (Python ' + \
                sys.version.split()[0] + ')'

        def do_POST(self):
            try:
                d = self.rfile.read(int(self.headers["content-length"]))
                if DEBUG:
                    print "\n***RECEIVING***\n", d, "*" * 13 + "\n"
                    sys.stdout.flush()
                (r,header) = parseSOAPRPC(d,1) # get header too

                method = r._name
                args   = r._aslist
                kw     = r._asdict

                ns = r._ns
                resp = ""
                # For faults messages
                if ns:
                    nsmethod = "%s:%s" % (ns, method)
                else:
                    nsmethod = method

                try:
                    # First look for registered functions
                    if self.funcmap.has_key(ns) and \
                       self.funcmap[ns].has_key(method):
                        f = self.funcmap[ns][method]
                    else: # Now look at registered objects
                        # Check for nested attributes
                        if method.find(".") != -1:
                            t = self.objmap[ns]
                            l = method.split(".")
                            for i in l:
                                t = getattr(t,i)
                            f = t
                        else:
                            f = getattr(self.objmap[ns], method)
                except:
                    resp = buildSOAP(Fault("%s:Client" % NS.ENV_T,
                       "No method %s found" % nsmethod,
                       "%s %s" % tuple(sys.exc_info()[0:2])),
                       encoding = self.encoding)
                    status = 500
                else:
                    try:
                        # If it's wrapped to indicate it takes keywords
                        # send it keywords
                        if header:
                            x = HeaderHandler(header)

                        if isinstance(f,MethodSig) and f.keywords:
                            tkw = {}
                            # This is lame, but have to de-unicode keywords
                            for (k,v) in kw.items():
                                tkw[str(k)] = v
                            fr = apply(f,(),tkw)
                        else:
                            fr = apply(f,args,{})
                        if type(fr) == type(self) and isinstance(fr, Void):
                            resp = buildSOAP(kw = {'%sResponse' % method:fr},
                                encoding = self.encoding)
                        else:
                            resp = buildSOAP(kw =
                                {'%sResponse' % method:{'Result':fr}},
                                encoding = self.encoding)
                    except Fault, e:
                        resp = buildSOAP(e)
                        status = 500
                    except:
                        resp = buildSOAP(Fault("%s:Server" % NS.ENV_T, \
                           "Method %s failed." % nsmethod,
                           "%s %s" % tuple(sys.exc_info()[0:2])),
                           encoding = self.encoding)
                        status = 500
                    else:
                        status = 200
            except Fault,e:
                resp = buildSOAP(e, encoding = self.encoding)
                status = 500
            except:
                # internal error, report as HTTP server error
                self.send_response(500)
                self.end_headers()
            else:
                # got a valid SOAP response
                self.send_response(status)
                self.send_header("Content-type",
                    'text/xml; charset="%s"' % self.encoding)
                self.send_header("Content-length", str(len(resp)))
                self.end_headers()

                if DEBUG:
                    print "\n***SENDING***\n", resp, "*" * 13 + "\n"
                    sys.stdout.flush()

                self.wfile.write(resp)
                self.wfile.flush()
                self.connection.shutdown(1)

    def __init__(self, addr = ('localhost', 8000),
        RequestHandler = SOAPRequestHandler, log = 1, encoding = 'UTF-8'):

        # Test the encoding, raising an exception if it's not known
        ''.encode(encoding)

        self.objmap  = {}
        self.funcmap = {}
        RequestHandler.encoding = encoding
        RequestHandler.objmap   = self.objmap
        RequestHandler.funcmap  = self.funcmap

        # Shut off logging
        if not log:
            def nop(*args):
                pass
            RequestHandler.log_message = nop

        self.allow_reuse_address = 1
        SocketServer.TCPServer.__init__(self,addr, RequestHandler)

    def registerObject(self, object, namespace=None):
        self.objmap[namespace] = object

    def registerFunction(self, function, namespace = None, funcName = None):
        if not funcName : funcName = function.__name__
        if self.funcmap.has_key(namespace):
            self.funcmap[namespace][funcName] = function
        else:
            self.funcmap[namespace] = {funcName : function}
    # convenience  - wraps your func for you.
    def registerKWFunction(self, function, namespace = None, funcName = None):
        self.registerFunction(MethodSig(function), namespace, funcName)
